/* UARTs/AT91/driver.s
AT91 UART driver

08/Mar/2000 ned
*/

#include "driver.h"

#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

static char modname[]="AT91UART";

typedef struct AT91UART_regs_t
{
	volatile int CR;			/* +0x00 */
	volatile int MR;			/* +0x04 */
	volatile int IER;			/* +0x08 */
	volatile int IDR;			/* +0x0c */
	volatile int IMR;			/* +0x10 */
	volatile int CSR;			/* +0x14 */
	volatile int RHR;			/* +0x18 */
	volatile int THR;			/* +0x1c */
	volatile int BRGR;			/* +0x20 */
	volatile int RTOR;			/* +0x24 */
	volatile int TTGR;			/* +0x28 */
	volatile int reserved0;		/* +0x2c */
	volatile int RPR;			/* +0x30 */
	volatile int RCR;			/* +0x34 */
	volatile int TPR;			/* +0x38 */
	volatile int TCR;			/* +0x3c */
} AT91UART_regs;

static void reset(AT91UART_regs *regs)
{
	int n;
	regs->CR=0xac;				/* Disable and reset transmitter */
	for(n=0; n<1024; n++);
	regs->RCR=regs->TCR=0;		/* Clear PDC counts (rx & tx) */
	regs->IDR=0xffffffff;		/* Disable ints */
}

extern HALError *AT91UARTDriver_Initialise(HALUARTBlk *blk, void *addr, int id)
{
	AT91UART_regs *regs;

	blk->UARTAddr=addr;
	regs=(AT91UART_regs *) addr;

	blk->id=id;
	blk->txbuff=0;
	blk->txbuffsize=0;
	blk->rxbuff=0;
	blk->rxbuffsize=0;

	return AT91UARTDriver_Set(blk, 9600, 8, 0, 1);
}

extern HALError *AT91UARTDriver_Finalise(HALUARTBlk *blk)
{
	AT91UART_regs *regs;
	s32 txsize;
	regs=(AT91UART_regs *) blk->UARTAddr;

	do
	{
		HALError *ret;
		if((ret=AT91UARTDriver_GetIOStatus(blk, &txsize, 0))!=0) return ret;
	} while(txsize);
	reset(regs);

	blk->UARTAddr=0;
	return 0;
}

extern HALError *AT91UARTDriver_Set(HALUARTBlk *blk, u32 baud, u8 bits, u8 parity, u8 stopbits)
{
	AT91UART_regs *regs;
	int n;
	regs=(AT91UART_regs *) blk->UARTAddr;

	reset(regs);

	blk->baud=baud;
	blk->bits=bits;
	blk->parity=parity;
	blk->stopbits=stopbits;

	/* Open rx & tx as outputs on PIO controller */
	*(volatile int *)(0xffff0004)=0x60c000;
	if(stopbits==1) stopbits=0;
	if(parity==0) parity=4;			/* none */
	else if(parity==1) parity=1;	/* odd */
	else if(parity==2) parity=0;	/* even */
	bits-=5;
	regs->MR=(stopbits<<12)+(parity<<9)+(bits<<6);	/* async */
	regs->BRGR=(int) ((32768000/16)/baud);
	regs->CR=0x50;					/* tx and rx enable */
	for(n=0; n<1024; n++);
	return 0;
}

extern HALError *AT91UARTDriver_WriteN(HALUARTBlk *blk, char *buff, int length)
{
	AT91UART_regs *regs;
	regs=(AT91UART_regs *) blk->UARTAddr;

	do
	{
		s32 txsize;
		do
		{
			HALError *ret;
			if((ret=AT91UARTDriver_GetIOStatus(blk, &txsize, 0))!=0) return ret;
		} while(txsize);
		regs->THR=*(buff++);
	} while(--length>0);
	return 0;
}

extern HALError *AT91UARTDriver_GetIOStatus(HALUARTBlk *blk, s32 *txsize, s32 *rxsize)
{
	static HALError e;
	AT91UART_regs *regs;
	int csr;
	regs=(AT91UART_regs *) blk->UARTAddr;

	csr=regs->CSR;
	if(txsize) *txsize=(csr & (1<<1)) ? 0 : -1;
	if(rxsize) *rxsize=(csr & (1<<0)) ? -1 : 0;

	/* Check for errors */
	if(csr & (1<<5))			/* Overrun */
	{
		memcpy(e.module.id, modname, 8);
		e.code=2;
		e.message="Data overrun";
		return &e;
	}
	if(csr & (1<<6))			/* Frame */
	{
		memcpy(e.module.id, modname, 8);
		e.code=3;
		e.message="Framing error";
		return &e;
	}
	if(csr & (1<<7))			/* Parity */
	{
		memcpy(e.module.id, modname, 8);
		e.code=4;
		e.message="Parity error";
		return &e;
	}
	if(csr & (1<<8))			/* Timeout */
	{
		memcpy(e.module.id, modname, 8);
		e.code=5;
		e.message="Timeout";
		return &e;
	}
	return 0;
}

extern HALError *AT91UARTDriver_ReadN(HALUARTBlk *blk, char *buff, int length)
{
	AT91UART_regs *regs;
	regs=(AT91UART_regs *) blk->UARTAddr;

	do
	{
		s32 rxsize;
		do
		{
			HALError *ret;
			if((ret=AT91UARTDriver_GetIOStatus(blk, 0, &rxsize))!=0) return ret;
		} while(!rxsize);
		*(buff++)=regs->RHR;
	} while(--length>0);
	return 0;
}

#ifdef __cplusplus
}
#endif

