;********************************************************************************************************
;                                               uC/OS-II
;                                         The Real-Time Kernel
;
;                        (c) Copyright 1992-1998, Jean J. Labrosse, Plantation, FL
;                                          All Rights Reserved
;
;
;                                           ARM Specific code
;
; File : OS_CPU_A.s
; By   : N. Douglas
;********************************************************************************************************

	AREA	OS_CPU_A, CODE, READONLY

	EXPORT	OSStartHighRdy
	EXPORT	OSCtxSw
	EXPORT	OSIntCtxSw
	EXPORT	OSTaskSw
	EXPORT	OSStackAndDisableInts
	EXPORT	OSRestoreStackedIntState

	IMPORT	OSTaskSwHook
	IMPORT	OSRunning
	IMPORT	OSTCBHighRdy
	IMPORT	OSTCBCur
	IMPORT	OSPrioHighRdy
	IMPORT	OSPrioCur

; Handy SWI definitions
OS_EnterOS	EQU	0x16

;*********************************************************************************************************
;                                          START MULTITASKING
;                                       void OSStartHighRdy(void)
;
; Assumes context is saved as:
; First R0-R12, R13_usr, R14_usr
; Then SPSR_irq (ie; CPSR_usr) and PC_usr
;*********************************************************************************************************

OSStartHighRdy
	; Need to be in a privilaged mode first
	SWI	OS_EnterOS

	; Call OSTaskSwHook()
	BL	OSTaskSwHook

	; Set OSRunning to TRUE
	MOV	R0,#1
	LDR	R1,=OSRunning
	STR	R0,[R1]
	
	; Get stack ptr for task to resume (lives in OSTCBHighRdy->OSTCBStkPtr)
	LDR	R1,=OSTCBHighRdy
	LDR	R1,[R1]
	LDR	R0,[R1,#0]				; OSTCBStkPtr is at +0

	; Restore saved context & return to task
	LDMFD	R0!,{R1,R14}				; SPSR & return addr
	MSR	SPSR_cxsf,R1
	LDMFD	R0,{R0-R14}^				; Restore USR mode regs R0-R14
	NOP
	MOVS	PC,R14					; PC=ret addr & CPSR=SPSR


;*********************************************************************************************************
;                                PERFORM A CONTEXT SWITCH (From task level)
;                                           void OSCtxSw(void)
;
;*********************************************************************************************************

OSCtxSw
	STMFD	R13!,{R14}				; Stack return addr

	; Save current task context
	STMDB	R13,{R13}^				; Determine R13_usr
	LDMDB	R13,{R14}
	STMFD	R14,{R0-R14}^				; Stack USR mode regs R0-R14
	MRS	R1, SPSR
	SUB	R0,R14,#14*4
	LDMFD	R13!,{R14}
	STMFD	R0!,{R1,R14}				; Stack SPSR & return addr

	; Store stack ptr of task (lives in OSTCBCur->OSTCBStkPtr)
	LDR	R1,=OSTCBCur
	LDR	R1,[R1]
	STR	R0,[R1,#0]				; OSTCBStkPtr is at +0

	; Call user definable OSTaskSwHook()
	BL	OSTaskSwHook

	; Set current task to new task
	LDR	R2,=OSTCBHighRdy
	LDR	R3,=OSTCBCur
	LDR	R4,=OSPrioHighRdy
	LDR	R5,=OSPrioCur
	LDR	R2,[R2]
	LDR	R3,[R3]
	LDR	R4,[R4]
	LDR	R5,[R5]
	LDR	R0,[R2]
	LDR	R1,[R4]
	STR	R0,[R3]
	STR	R1,[R5]

	; Get stack ptr for task to resume (lives in OSTCBHighRdy->OSTCBStkPtr)
	LDR	R1,=OSTCBHighRdy
	LDR	R1,[R1]
	LDR	R0,[R1,#0]				; OSTCBStkPtr is at +0

	; Restore saved context & return to task
	LDMFD	R0!,{R1,R14}				; SPSR & return addr
	MSR	SPSR_cxsf,R1
	LDMFD	R0,{R0-R14}^				; Restore USR mode regs R0-R14
	NOP
	MOVS	PC,R14					; PC=ret addr & CPSR=SPSR


;*********************************************************************************************************
;                                PERFORM A CONTEXT SWITCH (From an ISR)
;                                        void OSIntCtxSw(void)
;
;*********************************************************************************************************

OSIntCtxSw
	; Remove all registers on IRQ stack so far

	; Remove OSIntExit() stacked regs
	; {none stacked}

	; Remove ISR() stacked regs (KEEP SIMILAR)
	; {r4,r14}
	ADD	R13,R13,#2*4

	; Remove HAL stacked regs (KEEP SIMILAR)
	; {r0-r3,r9,r12,r14}
	LDMFD	R13!,{R0-R3,R9,R12,R14}

	; Perform normal task swap
	B	OSCtxSw


;*********************************************************************************************************
;                                        GENERATE SOFT INTERRUPT
;                                          void OSTaskSw(void)
;
;*********************************************************************************************************

OSTaskSw
	MOV	R2,R14
	MRS	R1,CPSR					; Take copy of current PSR
	SWI	OS_EnterOS				; enter SVC mode
	BIC	R0,R1,#0x1f
	ORR	R0,R0,#0x92				; IRQ32 + interrupts off
	MSR	CPSR_cxsf,R0
	NOP
	MSR	SPSR_cxsf,R1				; SPSR=old PSR
	MOV	R14,R2					; R14_irq=return addr
	MOV	PC,#0x18				; Branch to IRQ vector


;*********************************************************************************************************
;                                  STACK IRQ STATE AND DISABLE INTERRUPTS
;                                    void OSStackAndDisableInts(void)
;
; NOTE: Be careful calling from C code in SVC
;*********************************************************************************************************

OSStackAndDisableInts
	; Corrupts R0-R2 (okay during APCS use)
	MOV	R2,R14					; remember r14 (might be usr, irq, svc etc)
	SWI	OS_EnterOS				; enter SVC mode
	MSR	R0,SPSR					; disable interrupts in the SPSR
	ORR	R1,R0,#(1<<7)
	MRS	SPSR_cxsf,R1
	STMFD	R13!,{R0}				; stack the old SPSR on SVC stack
	MOVS	PC,R2					; CPSR=SPSR & PC=ret addr


;*********************************************************************************************************
;                                  STACK IRQ STATE AND DISABLE INTERRUPTS
;                                    void OSRestoreStackedIntState(void)
;
; NOTE: Be careful calling from C code in SVC
;*********************************************************************************************************

OSRestoreStackedIntState
	; Corrupts R0,R1 (okay during APCS use)
	MOV	R1,R14					; remember r14
	SWI	OS_EnterOS				; enter SVC mode
	LDMFD	R13!,{R0}				; load the old SPSR off SVC stack
	MRS	SPSR_cxsf,R0				; SPSR=what it did before
	MOVS	PC,R1					; CPSR=SPSR & PC=ret addr



	END